/****************************************************************************
  These are user-defined functions that can be bound to a Postgres backend
  and called by Postgres to execute SQL functions of the same name.

  The calling format for these functions is defined by the CREATE FUNCTION
  SQL statement that binds them to the backend.

  NOTE: These functions work with GUID (UUID)
*****************************************************************************
  Copyright (C) by AVN, 2007. All rights reserved.
  avn@avnsite.com
*/


# include <errno.h>
# include "postgres.h" /* general Postgres declarations */
# include "executor/executor.h"	/* for GetAttributeByName() */

# ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
# endif

/* These prototypes just prevent possible warnings from gcc */

Datum	uuid_bytea_to_char(PG_FUNCTION_ARGS);
Datum	uuid_char_to_bytea(PG_FUNCTION_ARGS);

/* By Value */

PG_FUNCTION_INFO_V1(uuid_bytea_to_char);

Datum uuid_bytea_to_char(PG_FUNCTION_ARGS)
{
  bytea *uuid_bytea = PG_GETARG_BYTEA_P( 0 );
  if( VARSIZE(uuid_bytea)-VARHDRSZ != 16 ) {
    ereport( ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
	     errmsg( "uuid_bytea_to_char: bytea(16) argument expected (VARSIZE=%d)",
	             VARSIZE(uuid_bytea)-VARHDRSZ)) );
  }
  BpChar *uuid_char = (BpChar *)palloc( 36+1+VARHDRSZ );
  if( uuid_char == NULL ) {
    ereport( ERROR, (errcode(ERRCODE_OUT_OF_MEMORY),
	     errmsg( "uuid_bytea_to_char: Can't allocate memory (errno=%d)", errno )) );
  }
  VARATT_SIZEP(uuid_char) = 36+VARHDRSZ;
  char *p = VARDATA(uuid_char);
  int i;
  unsigned char c;
  unsigned char *dat = (unsigned char *)VARDATA(uuid_bytea);
  for( i = 0; i < 16; i++ ) {
    if( /*+ optimization */ !(i&1) && (i == 4 || i == 6 || i == 8 || i == 10) )
      *p++ = '-';
    c = *dat >> 4;
    *p++ = ( c < 10 ? c+'0' : c-10+'a' );
    c = *dat++ & 0x0F;
    *p++ = ( c < 10 ? c+'0' : c-10+'a' );
  }
  *p = '\0';
//  elog( NOTICE, "%s", VARDATA(uuid_char) );
  PG_RETURN_BPCHAR_P( uuid_char );
}

PG_FUNCTION_INFO_V1(uuid_char_to_bytea);

Datum uuid_char_to_bytea(PG_FUNCTION_ARGS)
{
  BpChar *uuid_char = PG_GETARG_BPCHAR_P(0);

  if( VARSIZE(uuid_char)-VARHDRSZ != 36 ) {
    ereport( ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
	     errmsg( "uuid_char_to_bytea: char(36) argument expected (VARSIZE=%d)",
	             VARSIZE(uuid_char)-VARHDRSZ)) );
  }

  bytea *uuid_bytea = (bytea *)palloc( VARHDRSZ+16 );
  if( uuid_bytea == NULL ) {
    ereport( ERROR, (errcode(ERRCODE_OUT_OF_MEMORY),
	     errmsg( "uuid_char_to_bytea: Can't allocate memory (errno=%d)", errno )) );
  }

  VARATT_SIZEP(uuid_bytea) = VARHDRSZ+16;
  unsigned char *dst = (unsigned char *)VARDATA(uuid_bytea);
  char *src = VARDATA(uuid_char);
  int i, j;
  unsigned char c;
  for( i = j = 0; i < 16 && j < 36; i++ ) {
    if( /*+ optimization */ !(i&1) && (i == 4 || i == 6 || i == 8 || i == 10) ) {
      if( *src++ != '-' )
        ereport( ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                 errmsg( "uuid_char_to_bytea: invalid character '%c' detected at position %d",
		         *--src, j )) );
      j++;
    }

    if( *src >= '0' && *src <= '9' )
      c = (*src++ - '0') << 4;
    else if( *src >= 'A' && *src <= 'F' )
      c = (*src++ - 'A' + 10) << 4;
    else if( *src >= 'a' && *src <= 'f' )
      c = (*src++ - 'a' + 10) << 4;
    else
      ereport( ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
        errmsg( "uuid_char_to_bytea: invalid character '%c' detected at position %d", *src, j )) );
    j++;

    if( *src >= '0' && *src <= '9' )
      c |= (*src++ - '0');
    else if( *src >= 'A' && *src <= 'F' )
      c |= (*src++ - 'A' + 10);
    else if( *src >= 'a' && *src <= 'f' )
      c |= (*src++ - 'a' + 10);
    else
      ereport( ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
        errmsg( "uuid_char_to_bytea: invalid character '%c' detected at position %d", *src, j )) );
    j++;

    *dst++ = c;
  }
  if( j != 36 )
    ereport( ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
      errmsg( "uuid_char_to_bytea: invalid GUID string detected \"%s\"", VARDATA(uuid_char) )) );
  PG_RETURN_BYTEA_P(uuid_bytea);
}
